import unittest

from termtosvg.tests.test_anim import TestAnim
from termtosvg.tests.test_asciicast import TestAsciicast
from termtosvg.tests.test_config import TestConf
from termtosvg.tests.test_main import TestMain
from termtosvg.tests.test_term import TestTerm
